import pandas as pd
import json
import requests
from datetime import datetime
import logging
from tcia_utils.utils import searchDf
from tcia_utils.utils import remove_html_tags

_log = logging.getLogger(__name__)
logging.basicConfig(
    format='%(asctime)s:%(levelname)s:%(message)s'
    , level=logging.INFO
)

base_url = "https://cancerimagingarchive.net/api/v1/"

def getQuery(endpoint, per_page, format="", file_name=None, fields=None, ids=None, query=None, removeHtml=None):
    """
    Handle query basics that are common to all endpoints such as
    paging results, setting output formats, and file names.

    Args:
        endpoint (str): The API endpoint to query.
        per_page (int, optional): Number of results per page (default is 250).
        format (str, optional): Output format ('json' or 'df') (default is JSON if not populated).
        file_name (str, optional): File name to save the output as JSON or CSV if format = "df".
        ids (list, optional): List of IDs to include in the request (default is None).
        query (str, optional): Search criteria to filter results (default is None).

    Returns:
        list or DataFrame: Retrieved results based on the specified parameters and format.

    Raises:
        ValueError: If an invalid format is provided.

    """
    # Append custom fields to the endpoint if provided
    if fields:
        fields_str = ','.join(fields)
        endpoint += f"?_fields={fields_str}"
    
    # Set URL 
    url = base_url + endpoint
    
    # Set the request parameters
    params = {'per_page': per_page}
    
    # Add ids to the parameters if provided
    if ids:
        ids_str = ','.join(str(id) for id in ids)
        params['include'] = ids_str
    
    # Add query to the parameters if provided
    if query:
        params['search'] = query
    
    # Make a GET request to the API endpoint with the parameters
    _log.info('Requesting %s', url)
    response = requests.get(url, params=params)
    
    # Check if the request was successful
    if response.status_code == 200:
        # Parse the JSON response
        data = response.json()
        
        # Check if there are more pages to fetch
        while 'next' in response.links.keys():
            next_url = response.links['next']['url']
            _log.info('Requesting %s', next_url)
            response = requests.get(next_url)
            if response.status_code == 200:
                data.extend(response.json())
            else:
                _log.error('Error accessing the API: %s', response.status_code)
                break
        
        # Save or return the output based on the format
        if format == "json" or format == "":
            # Save as JSON
            if file_name:
                with open(file_name, "w") as f:
                    json.dump(data, f)
            return data
        elif format == "df":
            # Convert to DataFrame
            df = pd.DataFrame(data)
            # optionally remove HTML formatting for relevant columns
            if removeHtml == "yes":
                if "collections" in endpoint:
                    for column in ["collection_summary", "detailed_description", "publications_using", 
                                   "additional_resources", "collection_download_info", "publications_related",
                                   "version_change_log", "collection_acknowledgements"]:
                        if column in df:
                            df[column] = df[column].apply(remove_html_tags)
                elif "analysis" in endpoint:
                    for column in ["result_summary", "detailed_description", "publications_using", 
                                   "additional_resources", "collection_download_info", "publications_related",
                                   "version_change_log", "result_acknowledgements"]:
                        if column in df:
                            df[column] = df[column].apply(remove_html_tags)
                elif "downloads" in endpoint:
                    for column in ["description"]:
                        if column in df:
                            df[column] = df[column].apply(remove_html_tags)
                elif "citations" in endpoint:
                    for column in ["tcia_citation_text, tcia_citation_statement"]:
                        if column in df:
                            df[column] = df[column].apply(remove_html_tags)
                elif "versions" in endpoint:
                    for column in ["version_text"]:
                        if column in df:
                            df[column] = df[column].apply(remove_html_tags)

            # save csv if file name provided
            if file_name:
                df.to_csv(file_name, index=False)
            return df
        else:
            raise ValueError("Invalid format. Please choose 'json', 'df', or 'csv'.")
    else:
        _log.error('Error accessing the API: %s', response.status_code)
        return None


def getCollections(per_page=100, format="", file_name=None, fields=None, ids=None, query=None, removeHtml=None):
    """
    Retrieve collections from the API.

    Args:
        per_page (int, optional): Number of collections per page (default is 100).
        format (str, optional): Output format ('json' or 'df') (default is JSON if not populated).
        file_name (str, optional): File name to save the output as JSON or CSV if format = "df".
        fields (list, optional): List of custom fields to include in the return values (default is None).
                                 Possible fields: id, date, date_gmt, guid, modified, modified_gmt, slug, status,
                                 type, link, title, featured_media, template, yoast_head, yoast_head_json,
                                 cancer_types, citations, collection_doi, collection_download_info, collection_downloads,
                                 versions, additional_resources, cancer_locations, collection_page_accessibility,
                                 publications_related, version_change_log_archived, collection_status, publications_using,
                                 related_analysis_results, species, version_number, analysis_results, collection_title,
                                 date_updated, subjects, collection_short_title, data_types, detailed_description,
                                 version_change_log, collection_browse_title, supporting_data, collection_featured_image,
                                 collection_summary, collection_acknowledgements, collection_funding,
                                 hide_from_browse_table, program, _links.

    Returns:
        list or DataFrame: Retrieved collections based on the specified parameters and format.

    """
    # set the endpoint for Collections query
    endpoint = "collections/"
    
    # call getQuery to retrieve the data
    data = getQuery(endpoint, per_page, format, file_name, fields, ids, query, removeHtml)
    return data


def getAnalyses(per_page=100, format="", file_name=None, fields=None, ids=None, query=None, removeHtml=None):
    """
    Retrieve Analysis Results from the API.

    Args:
        per_page (int, optional): Number of analysis results per page (default is 100).
        format (str, optional): Output format ('json' or 'df') (default is JSON if not populated).
        file_name (str, optional): File name to save the output as JSON or CSV if format = "df".
        fields (list, optional): List of custom fields to include in the return values (default is None).
                                 Possible fields: id, date, date_gmt, guid, modified, modified_gmt, slug, status,
                                 type, link, title, featured_media, template, yoast_head, yoast_head_json,
                                 cancer_types, citations, result_doi, result_download_info, result_downloads,
                                 version_change_log_archived, versions, additional_resources, cancer_locations,
                                 publications_related, result_page_accessibility, detailed_description,
                                 publications_using, result_title, subjects, version_number, date_updated,
                                 related_collections, result_short_title, supporting_data, collections,
                                 result_browse_title, version_change_log, collection_downloads, result_summary,
                                 result_featured_image, result_acknowledgements, hide_from_browse_table, program,
                                 _links.

    Returns:
        list or DataFrame: Retrieved analysis results based on the specified parameters and format.

    """
    # set the endpoint for an Analysis Result query
    endpoint = "analysis-results/"
    
    # call getQuery to retrieve the data
    data = getQuery(endpoint, per_page, format, file_name, fields, ids, query, removeHtml)
    return data


def update_download_file_column(data):
    # Define a function to fetch 'source_url' from the API
    def fetch_source_url(row, current_url):
        if current_url:
            return current_url
        if isinstance(row, dict) and 'ID' in row:
            media_id = row['ID']
            response = requests.get(f'https://cancerimagingarchive.net/api/wp/v2/media/{media_id}')
            if response.status_code == 200:
                media_data = response.json()
                return media_data.get('source_url', '')
        return ''
    
    # Apply the function to the 'download_file' column to update it with the fetched URLs
    data['download_url'] = data.apply(lambda x: fetch_source_url(x['download_file'], x.get('download_url')), axis=1)
    return data


def getDownloads(per_page=200, format="", file_name=None, fields=None, ids=None, query=None, removeHtml=None):
    """
    Retrieve Download metadata from the API.

    Args:
        per_page (int, optional): Number of downloads per page (default is 200).
        format (str, optional): Output format ('json' or 'df') (default is JSON if not populated).
        file_name (str, optional): File name to save the output as JSON or CSV if format = "df".
        fields (list, optional): List of custom fields to include in the return values (default is None).
                                 Possible fields: id, date, date_gmt, guid, modified, modified_gmt, slug, status,
                                 type, link, title, template, yoast_head, yoast_head_json, cancer_type,
                                 download_file, download_requirements, download_size, download_title,
                                 cancer_location, data_license, download_size_unit, download_type, download_url,
                                 data_type, search_url, species, subjects, description, file_type, study_count,
                                 supporting_data, heading_example_text, series_count, download_access, image_count,
                                 collection_status, date_updated, _links.
        ids (list, optional): List of IDs to include in the request (default is None).
        query (str, optional): Search criteria to filter results (default is None).

    Returns:
        list or DataFrame: Retrieved download metadata based on the specified parameters and format.

    """
    # Set the endpoint for a Download query
    endpoint = "downloads/"
    
    # Call getQuery to retrieve the data
    data = getQuery(endpoint, per_page, format, file_name, fields, ids, query, removeHtml)
    
    if format == 'df':
        # Convert the data to a DataFrame if it's not already one
        if not isinstance(data, pd.DataFrame):
            data = pd.DataFrame(data)
        
        # Copy download URLs from download_file to download_url column
        data = update_download_file_column(data)
        
    return data


def getCitations(per_page=200, format="", file_name=None, fields=None, ids=None, query=None, removeHtml=None):
    """
    Retrieve Citation metadata from the API.

    Args:
        per_page (int, optional): Number of citations per page (default is 200).
        format (str, optional): Output format ('json' or 'df') (default is JSON if not populated).
        file_name (str, optional): File name to save the output as JSON or CSV if format = "df".
        fields (list, optional): List of custom fields to include in the return values (default is None).
                                 Possible fields: id, date, date_gmt, guid, modified, modified_gmt, slug, status,
                                 type, link, title, template, yoast_head, yoast_head_json, tcia_citation_type,
                                 tcia_citation_text, tcia_citation_statement, tcia_citation_doi, _links.
        ids (list, optional): List of IDs to include in the request (default is None).
        query (str, optional): Search criteria to filter results (default is None).

    Returns:
        list or DataFrame: Retrieved citation metadata based on the specified parameters and format.
    """
    # Set the endpoint for a Citation query
    endpoint = "citations/"
    
    # Call getQuery to retrieve the data
    data = getQuery(endpoint, per_page, format, file_name, fields, ids, query, removeHtml)
    return data


def getVersions(per_page=200, format="", file_name=None, fields=None, ids=None, query=None, removeHtml=None):
    """
    Retrieve Version metadata from the API.

    Args:
        per_page (int, optional): Number of citations per page (default is 200).
        format (str, optional): Output format ('json' or 'df') (default is JSON if not populated).
        file_name (str, optional): File name to save the output as JSON or CSV if format = "df".
        fields (list, optional): List of custom fields to include in the return values (default is None).
                                 Possible fields: id, date, date_gmt, guid, modified, modified_gmt, slug,
                                 status, type, link, title, template, yoast_head, yoast_head_json, 
                                 version_number, version_text, version_date, version_downloads, 
                                 related_collection, related_analysis_result, _links.
.
        ids (list, optional): List of IDs to include in the request (default is None).
        query (str, optional): Search criteria to filter results (default is None).

    Returns:
        list or DataFrame: Retrieved citation metadata based on the specified parameters and format.

    """
    # Set the endpoint for a Citation query
    endpoint = "versions/"
    
    # Call getQuery to retrieve the data
    data = getQuery(endpoint, per_page, format, file_name, fields, ids, query, removeHtml)
    return data


import pandas as pd
import inspect
from bs4 import BeautifulSoup
from unidecode import unidecode

def searchDf(search_term, dataframe='df', column_name=None):
    """
    This function searches for a term or a list of terms in a specified dataframe and column.
    
    Parameters:
    search_term (str or list): The term or list of terms to search for.
    dataframe (str or pd.DataFrame, optional): The dataframe to search in. If a string is provided, it is assumed to be the name of a dataframe in the global scope. Defaults to 'df'.
    column_name (str, optional): The name of the column to restrict the search to. If not provided, the search is performed across all columns.

    Returns:
    pd.DataFrame: A dataframe containing the rows where the search term was found.
    """
    import pandas as pd

    # If search_term is a string, convert it to a list
    if isinstance(search_term, str):
        search_term = [search_term]

    # If dataframe is a string, assume it's the name of a dataframe in the global scope
    if isinstance(dataframe, str):
        try:
            dataframe = globals()[dataframe]
        except KeyError:
            raise ValueError(f"No dataframe named '{dataframe}' found in the global scope.")

    # If column_name is provided, restrict the search to that column
    if column_name:
        try:
            contains_values = dataframe[column_name].apply(
                lambda x: any(value.lower() in str(x).lower() for value in search_term)
            )
        except KeyError:
            raise ValueError(f"No column named '{column_name}' found in the dataframe.")
    else:
        contains_values = dataframe.applymap(
            lambda x: any(value.lower() in str(x).lower() for value in search_term)
        )

    rows_with_values = contains_values if column_name else contains_values.any(axis=1)
    df_with_values = dataframe[rows_with_values]

    return df_with_values



def copy_df_cols(df_to_update, columns_to_copy, source_df, key_column):
    """
    Copy specified columns from source_df to df_to_update based on a key_column used as the lookup.
    
    Parameters:
    df_to_update (pd.DataFrame): The dataframe that needs to be updated.
    columns_to_copy (str or list): The column(s) to copy from source_df to df_to_update. Can be a single column name (str) or a list of column names.
    source_df (pd.DataFrame): The dataframe with the values you want to copy.
    key_column (str): The column name to use for matching the rows between the two dataframes.
    
    Returns:
    pd.DataFrame: The updated dataframe.
    """
    # If columns_to_copy is a string, convert it to a list
    if isinstance(columns_to_copy, str):
        columns_to_copy = [columns_to_copy]

    for column_to_copy in columns_to_copy:
        # Merging df_to_update with the specified column from source_df using the match_column as the key
        updated_df = df_to_update.merge(source_df[[key_column, column_to_copy]], 
                                        on=key_column, 
                                        how='left', 
                                        suffixes=('', '_lookup'))

        # Updating the specified column in df_to_update with the values from source_df
        updated_df[column_to_copy] = updated_df[f"{column_to_copy}_lookup"].combine_first(updated_df[column_to_copy])

        # Dropping the temporary column created during the merge
        updated_df.drop(columns=[f"{column_to_copy}_lookup"], inplace=True)

    return updated_df


def format_disk_space_binary(size_in_bytes):
    """
    Helper function for reportCollections() to format bytes to other binary units.
    I.e. Mebibytes (MiB) reported in Windows.
    """
    if size_in_bytes < 1024 ** 2:
        return f'{size_in_bytes / 1024:.2f} KB'
    elif size_in_bytes < 1024 ** 3:
        return f'{size_in_bytes / (1024 ** 2):.2f} MB'
    elif size_in_bytes < 1024 ** 4:
        return f'{size_in_bytes / (1024 ** 3):.2f} GB'
    elif size_in_bytes < 1024 ** 5:
        return f'{size_in_bytes / (1024 ** 4):.2f} TB'
    else:
        return f'{size_in_bytes / (1024 ** 5):.2f} PB'
        

def format_disk_space(size_in_bytes):
    """
    Helper function for reportCollections() to format bytes to other units.
    I.e. Megabytes (MB) reported in Mac/Linux.
    """
    if size_in_bytes < 1000:
        return f'{size_in_bytes:.2f} B'
    elif size_in_bytes < 1000 ** 2:
        return f'{size_in_bytes / 1000:.2f} kB'
    elif size_in_bytes < 1000 ** 3:
        return f'{size_in_bytes / (1000 ** 2):.2f} MB'
    elif size_in_bytes < 1000 ** 4:
        return f'{size_in_bytes / (1000 ** 3):.2f} GB'
    elif size_in_bytes < 1000 ** 5:
        return f'{size_in_bytes / (1000 ** 4):.2f} TB'
    else:
        return f'{size_in_bytes / (1000 ** 5):.2f} PB'
    

def remove_html_tags(text):
    """
    Helper function to convert HTML to plain text.
    """
    soup = BeautifulSoup(text, 'html.parser')
    plain_text = soup.get_text().strip()
    clean_text = unidecode(plain_text)  # Apply unidecode to remove or replace non-ASCII characters
    return clean_text